package main

import (
	"git.gammaspectra.live/S.O.N.G/Kirika/audio"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/filter"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format/aac"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format/flac"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format/mp3"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/format/opus"
	"git.gammaspectra.live/S.O.N.G/Kirika/audio/packetizer"
	"github.com/enriquebris/goconcurrentqueue"
	"io"
	"log"
	"sync"
)

type HeaderEntry struct {
	Name  string `json:"name"`
	Value string `json:"value"`
}

type ListenerInformation struct {
	Mount   string        `json:"mount"`
	Path    string        `json:"path"`
	Headers []HeaderEntry `json:"headers"`
}

type StreamListener struct {
	Information ListenerInformation
	Start       func(packets []packetizer.Packet) error
	Write       func(packet packetizer.Packet) error
	Close       func()
}
type StreamMount struct {
	Mount             string
	MimeType          string
	FormatDescription string
	Packetizer        packetizer.Packetizer
	Options           map[string]interface{}
	SampleRate        int
	Channels          int
	MetadataQueue     *goconcurrentqueue.FIFO
	listeners         []*StreamListener
	listenersLock     sync.Mutex
	keepBuffer        []packetizer.Packet
}

func NewStreamMount(source audio.Source, config *StreamConfig) *StreamMount {
	var encoderFormat format.Encoder
	options := make(map[string]interface{})
	var mimeType string

	reader, writer := io.Pipe()
	var packetizerEntry packetizer.Packetizer

	bitrate := config.GetOption("bitrate", nil)

	sampleRate := config.GetIntOption("samplerate", source.SampleRate)

	channels := config.GetIntOption("channels", source.Channels)

	switch config.Codec {
	case "vorbis":
		return nil
	case "opus":
		encoderFormat = opus.NewFormat()
		if bitrate != nil {
			options["bitrate"] = bitrate
		}
		sampleRate = opus.FixedSampleRate
		mimeType = "audio/ogg;codecs=opus"
		packetizerEntry = packetizer.NewOggPacketizer(reader)
	case "mp3":
		encoderFormat = mp3.NewFormat()
		if bitrate != nil {
			options["bitrate"] = bitrate
		}
		mimeType = "audio/mpeg;codecs=mp3"
		packetizerEntry = packetizer.NewMp3Packetizer(reader)
	case "flac":
		encoderFormat = flac.NewFormat()

		options["bitdepth"] = config.GetIntOption("bitdepth", 16)
		options["compression_level"] = config.GetIntOption("compression_level", 8)
		options["block_size"] = config.GetIntOption("block_size", 0)

		mimeType = "audio/flac"
		if config.Container == "ogg" {
			options["ogg"] = true
			mimeType = "audio/ogg;codecs=flac"
			packetizerEntry = packetizer.NewOggPacketizer(reader)
		} else {
			packetizerEntry = packetizer.NewFLACPacketizer(reader)
		}
	case "aac":
		encoderFormat = aac.NewFormat()
		if bitrate != nil {
			options["bitrate"] = bitrate
		}
		options["mode"] = config.GetStringOption("mode", "lc")
		mimeType = "audio/aac;codecs=mp4a.40.2"
		if options["mode"] == "he" || options["mode"] == "hev1" || options["bitrate"] == "vbr2" {
			mimeType = "audio/aac;codecs=mp4a.40.5"
		} else if options["mode"] == "hev2" || options["bitrate"] == "vbr1" {
			mimeType = "audio/aac;codecs=mp4a.40.29"
		}
		packetizerEntry = packetizer.NewAdtsPacketizer(reader)
	}

	if encoderFormat == nil {
		switch config.Container {
		case "ogg":
			encoderFormat = opus.NewFormat()
			if bitrate != nil {
				options["bitrate"] = bitrate
			}
			sampleRate = opus.FixedSampleRate
			mimeType = "audio/ogg;codecs=opus"
			packetizerEntry = packetizer.NewOggPacketizer(reader)
		case "mp3":
			encoderFormat = mp3.NewFormat()
			if bitrate != nil {
				options["bitrate"] = bitrate
			}
			mimeType = "audio/mpeg;codecs=mp3"
			packetizerEntry = packetizer.NewMp3Packetizer(reader)
		case "flac":
			encoderFormat = flac.NewFormat()

			options["bitdepth"] = config.GetIntOption("bitdepth", 16)
			options["compression_level"] = config.GetIntOption("compression_level", 8)
			options["block_size"] = config.GetIntOption("block_size", 0)

			mimeType = "audio/flac"
			packetizerEntry = packetizer.NewFLACPacketizer(reader)
		case "adts", "aac":
			encoderFormat = aac.NewFormat()
			if bitrate != nil {
				options["bitrate"] = bitrate
			}
			options["mode"] = config.GetStringOption("mode", "lc")
			mimeType = "audio/aac;codecs=mp4a.40.2"
			if options["mode"] == "he" || options["mode"] == "hev1" || options["bitrate"] == "vbr2" {
				mimeType = "audio/aac;codecs=mp4a.40.5"
			} else if options["mode"] == "hev2" || options["bitrate"] == "vbr1" {
				mimeType = "audio/aac;codecs=mp4a.40.29"
			}
			packetizerEntry = packetizer.NewAdtsPacketizer(reader)
		}
	}

	if encoderFormat == nil || packetizerEntry == nil {
		return nil
	}

	go func() {
		defer writer.Close()

		if channels != source.Channels {
			if channels == 1 {
				source = filter.MonoFilter{}.Process(source)
			} else if channels == 2 {
				source = filter.StereoFilter{}.Process(source)
			}
		}

		if sampleRate != source.SampleRate {
			source = filter.NewResampleFilter(sampleRate, filter.QualityFastest, 0).Process(source)
		}

		if err := encoderFormat.Encode(source, writer, options); err != nil {
			log.Panic(err)
		}
	}()

	return &StreamMount{
		Mount:             config.MountPath,
		MimeType:          mimeType,
		FormatDescription: encoderFormat.Description(),
		Packetizer:        packetizerEntry,
		SampleRate:        sampleRate,
		Channels:          channels,
		Options:           options,
		MetadataQueue:     goconcurrentqueue.NewFIFO(),
	}
}

func (m *StreamMount) removeDiscard(sampleNumber int64) {
	for i, p := range m.keepBuffer {
		if p.KeepMode() == packetizer.Discard && p.GetEndSampleNumber() <= sampleNumber {
			m.keepBuffer = append(m.keepBuffer[:i], m.keepBuffer[i+1:]...)
			m.removeDiscard(sampleNumber)
			break
		} else if p.GetEndSampleNumber() > sampleNumber {
			//they are placed in order
			break
		}
	}
}

func (m *StreamMount) removeKeepLast(category int64) {
	for i, p := range m.keepBuffer {
		if p.Category() == category && p.KeepMode() == packetizer.KeepLast {
			m.keepBuffer = append(m.keepBuffer[:i], m.keepBuffer[i+1:]...)
			m.removeKeepLast(category)
			break
		}
	}
}

func (m *StreamMount) removeGroupKeep(category int64) {
	for i, p := range m.keepBuffer {
		if p.Category() == category && p.KeepMode() == packetizer.GroupKeep {
			m.keepBuffer = append(m.keepBuffer[:i], m.keepBuffer[i+1:]...)
			m.removeGroupKeep(category)
			break
		}
	}
}

func (m *StreamMount) AddListener(listener *StreamListener) {
	m.listenersLock.Lock()
	defer m.listenersLock.Unlock()
	m.listeners = append(m.listeners, listener)
}

func (m *StreamMount) GetListeners() (entries []*ListenerInformation) {
	m.listenersLock.Lock()
	defer m.listenersLock.Unlock()
	for _, l := range m.listeners {
		entries = append(entries, &l.Information)
	}

	return
}

func (m *StreamMount) handlePacket(packet packetizer.Packet) {
	var toRemove []int
	//TODO: do this via goroutine messaging?
	for i, l := range m.listeners {
		if l.Start != nil {
			l.Start(m.keepBuffer)
			l.Start = nil
		}
		if l.Write(packet) != nil {
			toRemove = append(toRemove, i)
		}
	}

	if len(toRemove) > 0 {
		m.listenersLock.Lock()
		//TODO: remove more than one per iteration
		for _, i := range toRemove {
			l := m.listeners[i]
			m.listeners = append(m.listeners[:i], m.listeners[i+1:]...)
			l.Close()
			break
		}
		m.listenersLock.Unlock()
		toRemove = toRemove[:0]
	}

	sampleLimit := packet.GetEndSampleNumber() - int64(maxBufferSize*m.SampleRate)

	m.removeDiscard(sampleLimit) //always remove discards

	switch packet.KeepMode() {
	case packetizer.KeepLast:
		m.removeKeepLast(packet.Category())
		fallthrough
	case packetizer.Keep:
		m.keepBuffer = append(m.keepBuffer, packet)
	case packetizer.GroupKeep:
		m.keepBuffer = append(m.keepBuffer, packet)
	case packetizer.GroupDiscard:
		m.removeGroupKeep(packet.Category())
	case packetizer.Discard:
		m.keepBuffer = append(m.keepBuffer, packet)

	}
}

func (m *StreamMount) Process(group *sync.WaitGroup) {
	defer group.Done()
	defer func() {
		m.listenersLock.Lock()
		for _, l := range m.listeners {
			l.Close()
		}
		m.listeners = m.listeners[:0]
		m.listenersLock.Unlock()
	}()

	for {
		packet := m.Packetizer.GetPacket()
		if packet == nil {
			return
		}

		if item, err := m.MetadataQueue.Get(0); err == nil {
			if metadataPacket, ok := item.(*QueueMetadataPacket); ok {
				if packet.GetEndSampleNumber() > metadataPacket.GetStartSampleNumber() {
					m.MetadataQueue.Dequeue()
					m.handlePacket(metadataPacket)
				}

			}
		}

		m.handlePacket(packet)

	}

}
